/**************************************************************************
 *
 *        4-D File   ʸԡܶ ɤߤ պ Ѵؿ
 *            Created  by H.TSUKATA     1995
 *            Modified by H.Goto    Sep 1997
 *            Modified by H.Goto    Jul 2003
 *
 * Note: Some functions are now under construction.
 *
 **************************************************************************/

/*--------------------------------------------------------------------
  Copyright (C) 1994-2003  Hitoshi TSUKATA and Hideaki GOTO

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the authors' names, Hitoshi TSUKATA and Hideaki GOTO, may not be
  used in any advertising or otherwise to promote the sale, use or
  other dealings in this software without prior written authorization
  from the authors, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of modification is specified in cases
  where modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <limits.h>
#include <string.h>

// for C++ STL
#include <list>

#include "seg.h"
#include "line.h"

#define	h_kouryo( bp, ha)		((bp)->h >= (ha)/2.0 && (bp)->h <= (ha)*2 )
#define	slope_kouryo( slp, slp_ave)	( fabs( (slp)-(slp_ave))<2.0)

extern OPTION	opt;


/*--------------------------------------------------------------
    Byte order exchange
--------------------------------------------------------------*/

short byte_ch_short( short src){
  short ret;
  ret  = ((unsigned short)src >> 8) & 0x00ff;
  ret |= ((unsigned short)src << 8) & 0xff00; 
  return ret;
}


/*--------------------------------------------------------------
    Endian-controled I/O
--------------------------------------------------------------*/

static int read_bigshort_array(FILE *fp,short *d,int n){
	int	i;
	unsigned char	buf[2];
	for ( i=0 ; i<n ; i++ ){
		if ( 1 != fread((void *)buf,2,1,fp) )  return(-1);
		*d  = (short)((unsigned short)buf[0] << 8);
		*d |= (short)((unsigned short)buf[1]);
		d++;
	}
	return(0);
}


static int write_bigshort_array(FILE *fp,short *d,int n){
	int	i;
	unsigned char	buf[2];
	for ( i=0 ; i<n ; i++ ){
		buf[0] = (unsigned char)((unsigned short)*d >> 8);
		buf[1] = (unsigned char)*d;
		d++;
		if ( 2 != fwrite((void *)buf,1,2,fp) )  return(-1);
	}
	return(0);
}


/*--------------------------------------------------------------
    D4 file Primitive I/O
--------------------------------------------------------------*/

#ifndef d4_BUFFERS
#define		d4_BUFFERS	1024
#endif

#ifndef D4_MAGIC
#define	D4_MAGIC		"DATA0004" 
#endif


class d4file {
  /*	D4եΥإå		
   */
public:
  char		d4_magic[8];
  long		d4_attrb;
  long		d4_count;
  short	d4_llimit[4];
  short	d4_hlimit[4];
  short	d4_spc[48];

  d4file(void){
	int	i;
	strcpy( d4_magic, D4_MAGIC);
	d4_attrb = 0;
	d4_count = 0;
	for(i=0; i<4; i++){
	  d4_hlimit[i]=0;
	  d4_llimit[i]=0;
	}
	for(i=0; i<48; i++ )	d4_spc[i]=0;
  }

  int readheader(FILE *fp);
  int writeheader(FILE *fp);
};


int d4file :: readheader(FILE *fp){
	unsigned short	d[2];
	if ( 1 != fread(d4_magic,8,1,fp) ){
		fprintf( stderr,"\n bad magic numger - not a D4-format file\n");
		exit(1);
	}
	if ( 0 != strncmp(d4_magic,D4_MAGIC,8) ){
		fprintf( stderr,"\n bad magic numger - not a D4-format file\n");
		exit(1);
	}
	if ( read_bigshort_array(fp,(short *)d,2) ){
		fprintf( stderr,"\n bad attribute - not a D4-format file\n");
		exit(1);
	}
	d4_attrb = (long)(((unsigned long)d[0] << 16) | (unsigned long)d[1]);
	if ( d4_attrb != 0 ){
		fprintf( stderr,"\n bad attribute - not a D4-format file\n");
		exit(1);
	}
	if ( read_bigshort_array(fp,(short *)d,2) ){
		fprintf( stderr,"\n bad count - not a D4-format file\n");
		exit(1);
	}
	if ( read_bigshort_array(fp,d4_llimit,4) ){
		fprintf( stderr,"\n bad header - not a D4-format file\n");
		exit(1);
	}
	if ( read_bigshort_array(fp,d4_hlimit,4) ){
		fprintf( stderr,"\n bad header - not a D4-format file\n");
		exit(1);
	}
	if ( read_bigshort_array(fp,d4_spc,48) ){
		fprintf( stderr,"\n bad header - not a D4-format file\n");
		exit(1);
	}
	return(-1);
}


int d4file :: writeheader(FILE *fp){
	unsigned short	d[2];
	int	i;
	if ( 1 != fwrite(D4_MAGIC,8,1,fp) ){
		fprintf( stderr,"\n write error on D4-file\n");
		exit(1);
	}
	d[0] = d[1] = 0;
	if ( write_bigshort_array(fp,(short *)d,2) ){
		fprintf( stderr,"\n write error on D4-file\n");
		exit(1);
	}
	d[0] = d[1] = 0;
	if ( write_bigshort_array(fp,(short *)d,2) ){
		fprintf( stderr,"\n write error on D4-file\n");
		exit(1);
	}
	if ( write_bigshort_array(fp,d4_llimit,4) ){
		fprintf( stderr,"\n write error on D4-file\n");
		exit(1);
	}
	if ( write_bigshort_array(fp,d4_hlimit,4) ){
		fprintf( stderr,"\n write error on D4-file\n");
		exit(1);
	}
	for ( i=0 ; i<48 ; i++ )  d4_spc[i] = 0;
	if ( write_bigshort_array(fp,d4_spc,48) ){
		fprintf( stderr,"\n write error on D4-file\n");
		exit(1);
	}
	return(-1);
}


class d4pac {
public:
  short		d4_data[4];
  int allzero( void){
	return( d4_data[0]==0 && d4_data[1]==0 &&
			d4_data[2]==0 && d4_data[3]==0 );
  }
  d4pac(void){
	d4_data[0]=0; d4_data[1]=0; d4_data[2]=0; d4_data[3]=0;
  }
	
  void byte_ch(void);
};


void d4pac::byte_ch(void){
	int i;
	d4pac b = *this;
	for( i=0; i<4; i++)
	  d4_data[i]= byte_ch_short( b.d4_data[i]);
}


/*--------------------------------------------------------------
    D4 file I/O
--------------------------------------------------------------*/

d4pac * d4_readd4( FILE *fp, int *num){
  d4pac *ret;
  d4file	hd;
  std::list<d4pac *>	buf;
  d4pac	tmp, *tmp2;
  int	i;
  // إå
  hd.readheader(fp);
  // 
  for( buf.clear(); fread( &tmp, sizeof( d4pac), 1, fp)!=0; ){
#ifdef LSB_FIRST
	tmp.byte_ch();
#endif
	tmp2 = new d4pac(tmp);
	buf.push_back(tmp2);
  }  
  //ܤ
  ret = new d4pac[ buf.size()];
  for( i=0; ! buf.empty() ; buf.pop_front(), i++){
	ret[i] = *(buf.front());
	delete buf.front();
  }
  *num = i;
  return ret;
}


#if 0
void d4_writed4( FILE *fp, d4pac *src ,int num){
  int	i;
  d4file	hd;
  // إå
  strcpy( hd.d4_magic, D4_MAGIC);

  /* set header information here */




  /* -------- */

  hd.writeheader(fp);

  // 
  for ( i=0 ; i<num ; i++ ){
     write_bigshort_array(fp,src[i].d4_data,4);
  }
}
#endif




/*--------------------------------------------------------------
    Rectangle class
--------------------------------------------------------------*/

int Rectangle::chk_marker( void){
  /*
   *	ޡΥå
   */
  if( x==0 && y==0 && w==0 && h==0)		return (1);	// ʸԽüޡ
  if( x==-1 && y==-1 && w==0 && h==0)	return (2);	// ʸϥ֥åüޡ
  return (0);			// ޡȤϴطʤ󤸤㡼
}

void Rectangle::set_marker( short flag=0){
	/*
	 *	ޡΥå	
	 *		flag	0;ʸԽüޡ 
	 *				1;ʸϥ֥åüޡ
	 */
  if( flag){
	x=-1; y=-1; w=0; h=0;		// ʸϥ֥åüޡ
  } else {
	x=0; y=0; w=0; h=0;		// ʸԽüޡ
  }
}



int spcdata( LSEG *sgp){
  return( sgp->h<=0 && sgp->w<=0);
}

int spcdata( lseginfo *slp){
  return( slp->htpeak == 0);
}


void Rectangle::byte_ch(void){
	Rectangle b;
	b = *this;
	x = byte_ch_short( b.x);
	y = byte_ch_short( b.y);
	w = byte_ch_short( b.w);
	h = byte_ch_short( b.h);
}

double slope_av( STRING *sp0, STRING *sp1){
  /* 
   *   =ξԤηΡŤߡդʿ 
   */
  double ret;
  ret = deg2rad( (sp0->slope*sp0->w+sp1->slope*sp1->w) / (double)(sp0->w+sp1->w));
  return(ret);
}



int str_overlap( STRING *s0, STRING *s1){
  double theta =  slope_av( s0, s1);
  double	_x[4],_y[4];
  _x[0]= s0->x; 	_y[0]=s1->y; 

  _x[1]= ( s0->x+s0->w + _x[0]*tan(theta)*tan(theta) + 
		   ( s0->y-s0->w*sin(deg2rad(s0->slope)) - _y[0]) * tan(theta))
	/ ( 1 + tan(theta)*tan(theta) );
  _y[1]=tan(theta)*(_x[0]-_x[1]) + _y[0];

  _x[2]= ( s1->x + _x[0]*tan(theta)*tan(theta) + 
		   ( s1->y - _y[0]) * tan(theta))
	/ ( 1 + tan(theta)*tan(theta) );
  _y[2]=tan(theta)*(_x[0]-_x[2]) + _y[0];
	
  _x[3]= ( s1->x+s1->w + _x[0]*tan(theta)*tan(theta) + 
		   ( s1->y-s1->w*sin(deg2rad(s1->slope)) - _y[0]) * tan(theta))
	/ ( 1 + tan(theta)*tan(theta) );
  _y[3]=tan(theta)*(_x[0]-_x[3]) + _y[0];


  double ax,ay, bx,by;
  if( _x[0] <  _x[2]){	ax = _x[2]; ay=_y[2];
  } else {				ax = _x[0]; ay=_y[0];  }

  if( _x[1] <  _x[3]){	bx = _x[1]; by=_y[1];
  } else {				bx = _x[3]; by=_y[3];  }
  
  double ret = ((ax<bx) ? 1 : -1)* sqrt( (bx-ax)*(bx-ax) + (by-ay)*(by-ay));
  return ( (int)ret);
}

int rect_overlap( Rectangle *b0, Rectangle *b1){
  return ( min( b0->x+b0->w, b1->x+b1->w) - max( b0->x, b1->x));
}

void puttlterm( FILE *fp){
  /*
   *	ʸԤνޡ
   */
  short	marker[4] = {0,0,0,0};
  write_bigshort_array(fp,marker,4);
#if 0
  Rectangle	b(0,0,0,0);
  fwrite( &b, sizeof( Rectangle), 1, fp);
#endif
}
  
void getspcdata( Rectangle *bp){
  bp->h = 0;
  bp->w = 0;
}

double degree_angle(double x, double y){

  double rad,r;
  r = sqrt( x*x + y*y);
  rad = acos( x / r);
  if ( asin( y / r) < 0.0 )
	rad = -rad;

  return ( rad * 180.0 / PI) ;
}

Rectangle * read_cr( FILE *fp, int *cnt){
  d4pac	*d4;
  int	i;
  Rectangle	*ret;

  //		ö d4pac ɤ,Rectangle ˰ư
  d4 = d4_readd4( fp, cnt);
  ret = new Rectangle[ *cnt+1]();
  for( i=0; i<*cnt; i++){
	ret[i].x = d4[i].d4_data[0];
	ret[i].y = d4[i].d4_data[1];
	ret[i].w = d4[i].d4_data[2];
	ret[i].h = d4[i].d4_data[3];
  }
  delete d4;	// d4pac,
  return ret;
}




/*--------------------------------------------------------------
    STRING class
--------------------------------------------------------------*/

int STRING::readdata( FILE *fp){
  /*	ʸԤߤ */
  Rectangle	btmp, *btmp1;		
  std::list<Rectangle*>	buf;
  int	i;

  buf.clear();
  while( fread( &btmp, sizeof( Rectangle), 1, fp)!=0 ){
#ifdef LSB_FIRST
	btmp.byte_ch();
#endif
	btmp1 = new Rectangle( btmp);
	buf.push_back(btmp1);
	if( btmp.chk_marker() == 1)
	  break;
  }
  if( (i = buf.size()) == 0 ) return 0;
  box = new Rectangle[i];
  for( i=0; ! buf.empty() ; buf.pop_front(), i++){
	box[i] = *(buf.front());
	delete buf.front();
  }
  calc_data();	//׻⤹뤸硼
  box_num = i-2;
  return( i-1);		//   i ϥޡޤ᤿Ŀ
}


void STRING::writedata( FILE *fp){
  Rectangle	*bp;
  bp=box; 
  do{
	write_bigshort_array(fp,(short *)bp, 4);
  }while( (bp++)->chk_marker()==0);
}

  
void STRING::calc_data( void){
  /*
   *	ʸԥǡη׻
   *
   */
	
  Rectangle		*bp ;
  int	n;
  int  jn;		/* jn -- θоݤȤθĿ */ 
  LINEHEIGHT sum_h, h_ave_tmp;

  // ⤵Ρֲʿѡפ
  for( bp=box,sum_h=0.0,n=0; bp->chk_marker()==0; bp++,n++){
	sum_h +=(double) bp->h;
  }
  h_ave_tmp = (LINEHEIGHT)(sum_h / (double)(n)); /*Ȥꤢι⤵ʿ*/

  // Ǿˡˤ극Ф
  double	m, b; 	// y=mx+b
  double	sum_x, sum_y, sum_xx, sum_yy, sum_xy;  //x, y, x^2, y^2, xy
  double	mx, my;
  for( bp=box,sum_x=0.0,sum_y=0.0,sum_xx=0.0,sum_yy=0.0,sum_xy=0.0,n=0;
	  bp->chk_marker()==0; bp++){
	mx = bp->x+bp->w/2; 	my = bp->y+bp->h/2;  // 濴ɸ
	if( !((bp==box||(bp+1)->chk_marker()!=0) && bp->h<h_ave_tmp*0.3) ){
	  //ξüζ,⤵ʿѤι⤵㤹,̵
	  sum_x += mx;
	  sum_y += my;
	  sum_xx += mx * mx;
	  sum_yy += my * my;
	  sum_xy += mx * my;
	  n++;
	}
  }
  // x,y,w,slope  Ƿ
  x = box->x;
  w = (bp-1)->x+(bp-1)->w - box->x;
  if ( n*sum_xx != sum_x*sum_x){
	m = (n*sum_xy -sum_x*sum_y)/(n*sum_xx-sum_x*sum_x);
	b = (sum_x*sum_xy-sum_xx*sum_y)/(sum_x*sum_x-n*sum_xx);
	slope = rad2deg( atan( -m));
	y = (int)(m*(double)x + b);	//y=mx+b  
  } else {			// ˤΤ,ͭ1ĤΤȤ(..Ȼפ)
	slope = 0.0;
	y = (int)( sum_y/(double)n );
  }
  
  /* Height Check  */
  //   ܶΤ,ؤ㤹Τֹ⤵פ
  //   θǡƹ⤵׻.
  sum_h=0, sum_y=0;
  for(  bp=box ;
		bp->chk_marker()==0 && h_kouryo(bp, h_ave_tmp)!=1 ;
		bp++);	
  //			⤵ͰʾȤʤǡ
  //			ޤǥݥ󥿤ʤ. 
  sum_h += bp->h;   sum_y += bp->y + bp->h/2.0; 
  for( jn=1, bp++ ; bp->chk_marker()==0; bp++){
	if( h_kouryo(bp, h_ave_tmp)){
	  sum_h += bp->h ; sum_y += bp->y + bp->h/2.0;
	  jn++;
	}
  }
  h = (double)sum_h / (double)jn;

  /* νſɸ */
  cx = x + w / (double)2 ;
  cy = y - w / (double)2 * tan( deg2rad( slope));
}




/*--------------------------------------------------------------
    Document class
--------------------------------------------------------------*/

void  Document::writelinedata( FILE *fp){
  /*
   *	ʸԷ ǡ 
   */
  STRING	*sp;
  Rectangle	*bp;
  d4file	hd;
  int	i;

  // إåΥǡå
  hd.d4_count =0;
  for( i=0; i<4; i++){
	hd.d4_llimit[i] = SHRT_MAX;
	//	hd.d4_llimit[i] =  0;
	hd.d4_hlimit[i] =  0;
  }
  for( sp=line; sp->chk_marker()==0; sp++){
	for( bp=sp->box; bp->chk_marker()==0; bp++){
	  //	  hd.d4_count ++;
	  hd.d4_llimit[0] = min( bp->x, hd.d4_llimit[0]);
	  hd.d4_llimit[1] = min( bp->y, hd.d4_llimit[1]);
	  hd.d4_llimit[2] = min( bp->w, hd.d4_llimit[2]);
	  hd.d4_llimit[3] = min( bp->h, hd.d4_llimit[3]);
	  hd.d4_hlimit[0] = max( bp->x, hd.d4_llimit[0]);
	  hd.d4_hlimit[1] = max( bp->y, hd.d4_llimit[1]);
	  hd.d4_hlimit[2] = max( bp->x+bp->w, hd.d4_llimit[2]);
	  hd.d4_hlimit[3] = max( bp->y+bp->h, hd.d4_llimit[3]);
	}
  }
  hd.writeheader(fp);

  // ʸǡ񤭤
  sp = line;
  while( sp->chk_marker()==0){
	(sp++)->writedata( fp);
  };
  
}


int Document::readlinedata( FILE *fp){
  int			line_cnt=0;
  STRING		*sp;
  std::list<STRING*>	buf;
  d4file		*hd;
  
  // إå
  hd = new d4file();
  hd ->readheader( fp);

  // 
  buf.clear();
  sp = new STRING;
  while( sp->readdata(fp)>0){
	buf.push_back(sp);
	sp = new STRING();
  }
  sp->set_marker();
  buf.push_back(sp);

  // ܤ ( Ϣ³ΰؤγǼΤɬ)
  line = new STRING[buf.size()]; 
  for( line_cnt=0; ! buf.empty() ; buf.pop_front(), line_cnt++){
	line[line_cnt] = *(buf.front());
	line[line_cnt].index = line_cnt;
	delete buf.front();	// ʤߤܤΤ,ХåեѤʤ
  }
  line_num = line_cnt-1;
  return ( line_cnt-1);
}


int compare( STRING *sp0, STRING *sp1){
  /*
   *	ʸԤΡְ֡״ط
   *				+1; sp0 
   *					-1; sp1 					
   *					 0; ξԤƱ
   */
  if( fabs(sp0->cy-sp1->cy) > (sp0->h+sp1->h)*.5*.3)
	return ( (sp0->cy < sp1->cy) ? 1 : -1);
  else
	return ( (sp0->cx < sp1->cx) ? 1 : -1);
  return(0);
} 
	  

double linedistance(STRING *sp0, STRING *sp1)
{
  /*
   *   n0  n1 ֲ <=> ()>0
   *
   */
  double       theta,d;// x1,y1, x2, y2;

  /*
   *   ֿפνſ
   */
  //  x1 = sp0->cx;   y1 = sp0->cy;
  //  x2 = sp1->cx;   y2 = sp1->cy;
   
//   // *ſ*濴˲ž  **濴˲ ž
//   x1 = sp0->x;
//   x2 = sp1->x;
//   y1 = sp0->y;
//   y2 = sp1->y;

  theta = slope_av( sp0, sp1);

  //  x1 = sp0->cx;   y1 = sp0->cy;
  //  x2 = sp1->cx;   y2 = sp1->cy;
  d = (sp1->cy - sp0->cy) * cos(theta) 	
	+ (sp1->cx - sp0->cx) * sin(theta);
  
  return ( d);
}
       

void makenss(STRING *line)
{
  /*
   *   ˵ʸԽ (Neighboring Strings Set) 
   */

  double dist;
  STRING	*sp, *sp1;

  for ( sp=line; sp->chk_marker()==0; sp++){ 
	sp->adj.clear();
	for ( sp1=line; sp1->chk_marker()==0; sp1++){
	  if ( sp1 != sp){
		dist = linedistance( sp, sp1);
		if ( fabs(dist) < sp->h*T4 && hi((double)sp->h, (double)sp1->h) < 1.4)
		  sp->adj.push_back( sp1);
	  }
	}
  }
}


void makens(STRING *line){
  /*
   *   Ƕʸ (Nearest Srting)   
   *   * * θ * *
   */
   
  double	dist, takasa_hi, kasanari, haba_hi;
  double	nrst_dist[2];
  STRING	*nrst_node[2];
  int		i;
  int		nrst_dir ;
  STRING	*sp, *sp1;

  for ( sp=line; sp->chk_marker()==0; sp++){

	nrst_node[0]=NULL ; nrst_node[1]=NULL;
	nrst_dist[0]=MAX_DIST ;  nrst_dist[1]=MAX_DIST ;
	for ( sp1=line; sp1->chk_marker()==0; sp1++){
	  if ( sp1 != sp){
		dist = linedistance( sp, sp1);/* ܹԤ־ */
		kasanari = str_overlap( sp, sp1);
		
		if (fabs(dist) < fabs( nrst_dist[(dist>0)])      /* Ƕܴط */
			//			&& kasanari > min( sp->w, sp1->w) * .1      /* Ťʤ */
			&& kasanari > T61*( min( sp->w, sp1->w))      /* Ťʤ */
			&& fabs(dist)< sp->h*4.5      /* ==ٶ᤯ʤȥ */
			){
		  nrst_dist[(dist>0)] = dist ;
		  nrst_node[(dist>0)] = sp1 ;
		}
	  }
	}


	/* & ⤵å */
	if(  min(nrst_dist[0], nrst_dist[1])< MAX_DIST)
	  nrst_dir = ( fabs(nrst_dist[0]) < fabs(nrst_dist[1])) ? 0 : 1 ;
	else
	  nrst_dir = EMPTY_TREE;
      
	if( nrst_dir!=EMPTY_TREE){
	  haba_hi = hi( (double)sp->w, (double) nrst_node[nrst_dir]->w  );
	  takasa_hi = hi((double)sp->h, (double) nrst_node[nrst_dir]->h );
	  if( haba_hi>T60 || takasa_hi>T5 ) {
		// ,⤵椬ԹʤǤ ⤦¾Ƥɶᤱ
		// äθ
		if( hi( fabs( nrst_dist[0]), fabs(nrst_dist[1])) < 1.2){	
		  haba_hi = hi( (double)sp->w, (double) nrst_node[1-nrst_dir]->w  );
		  takasa_hi = hi((double)sp->h, (double) nrst_node[1-nrst_dir]->h);
		  if( haba_hi <= T60 && takasa_hi <= T5)
			nrst_dir = 1-nrst_dir;
		  else
			nrst_dir = EMPTY_TREE;
		} else{
		  nrst_dir = EMPTY_TREE;
		}
	  } 
	}

	if( nrst_node[nrst_dir] != NULL && nrst_dir!=EMPTY_TREE){
	  sp->adj.push_back( nrst_node[nrst_dir]);	// ***ϤΤä 0000
//	  sp->nd_adj_node.push_back( nrst_node[nrst_dir]);	//ܥΡɥꥹȤɲ
	  
	}
	for( i=0 ; i<2 ; i++){
	  sp->nearestnode[i] = nrst_node[i];
	  sp->nearestdist[i] = nrst_dist[i];
	}
  }
}
