/*----------------------------------------------------------------------
    PBM File I/O
	Created  by H.TSUKATA       1995
        Modified by H.Goto    Sept. 1997
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1995-1997  Hitoshi TSUKATA

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hitoshi TSUKATA, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>

#include "ufilep.h"
#include "pbm_file.h"

#undef min
#define min(a,b) ((a) < (b) ? (a) : (b))


/*  C Part follows  */

bit** pbm_allocarray(int cols,int rows){
	bit	**t;
	int	i;
	if ( NULL == (t = (bit**)malloc(cols * rows * sizeof(bit*))) )  return(NULL);
	if ( NULL == (t[0] = (bit*)malloc(cols * rows * sizeof(bit))) ){
		free((void*)t);
		return(NULL);
	}
	for ( i=1 ; i<rows ; i++ )  t[i] = t[0] + (i * cols);
	return(t);
}


void pbm_freearray(bit **t,int rows){
	if ( NULL == t )  return;
	free((void*)t[0]);
	free((void*)t);
}


/*  C++ Part follows  */

void pbm_file::pbm_init(void){
  if( width>0 && height>0)
	pbm_freearray( bits, width);
  width = 0;	height = 0;	
}
	

int pbm_file::readpbm(char *fname){
	PBMFILE	PBMLD;
	int	i,retcode;
	if ( PBMLD.open(fname,"r") )  return(-1);
	if ( PBMLD.getfiletype() != 0 ){  PBMLD.close();  return(-1); }
	width = PBMLD.getwidth();
	height = PBMLD.getheight();
	PBMLD.setpal_fb(1,0);
	if ( NULL == (bits = pbm_allocarray(width,height)) ){
		PBMLD.close();  return(-1);
	}
	retcode = 0;
	for ( i=0 ; i<height ; i++ ){
		if ( PBMLD.readline_gray(-1,(void *)bits[i]) ){
			retcode = -1;  break;
		}
	}
	PBMLD.close();
	return(retcode);
}

int pbm_file::writepbm(char *fname){
	PBMFILE	PBMSV;
	int	i,retcode;
	PBMSV.setsize(width,height,0);
	if ( PBMSV.open(fname,"w") )  return(-1);
	retcode = 0;
	for ( i=0 ; i<height ; i++ ){
		if ( PBMSV.writeline_bilevel(-1,(void *)bits[i],1) ){
			retcode = -1;  break;
		}
	}
	if ( PBMSV.close() )  retcode = -1;
	return(retcode);
}

void pbm_file::pbm_alloc( int w, int h){
  int	i,j;
  width = w;
  height = h;
  bits = pbm_allocarray(  width, height);
  //Ƥ 0 
  for( i=0; i<w; i++)
	for( j=0; j<h; j++)	
	  bits[j][i] = 0;
}

void pbm_file::pbm_free( void){
  pbm_freearray( bits, height);
	
}

void pbm_file::pbm_crop( pbm_file & n_img,int n_col, int n_row, int n_cols, int n_rows){
  int		i,j;
		
  n_img.pbm_init();
  n_img.pbm_alloc(  n_cols, n_rows);
  for( i=n_col; i<min( n_col+n_cols,  width); i++)
	for( j=n_row; j<min(n_row+n_rows,height); j++)
	  n_img.bits[j-n_row][i-n_col]= bits[j][i];	
}


  
int pbm_file::pbmprojection(int **prj, int flg){
   
  //      ᡼ΥҥȥȤ.
  //		
  //			flg == 1; ʿҥȥ 
  //			ʳ; ľҥȥ

  int		i,j;
  int		prjtmp;	
  int		ax1, ax2;
   
  ax1 = ( flg==1) ? height : width ; 
  ax2 = ( flg!=1) ? height : width ; 

  (*prj)=(int *)calloc( ax1+1, sizeof( int));
  for( i=0 ; i < ax1 ; i++){ 
	prjtmp=0;
	for ( j=0 ; j < ax2 ; j++){
	  prjtmp += (flg!=1) ? bits[j][i]: bits[i][j] ;
	}
	(*prj)[i]=prjtmp;
  }
  (*prj)[ i] = -1;	//ҥȥκǽü -1
	  //	  fclose(fp);
	  return( 0);
}


void pbm_file::pbmfold( int v){

  int	i,j;
  bit	buf;
  
  if(v){
	/*  ꤫ = ľ */
	for( i=0; i<width/2; i++){
	  for( j=0; j<height; j++){
		buf = bits[j][i];
		bits[j][i] = bits[j][ width-1-i];
		bits[j][ width-1-i] = buf;
	  }
	}
  }else{
	/*  ꤫ = ʿ*/
	for( j=0; j<height/2; j++){
	  for( i=0; i<width; i++){
		buf = bits[j][i];
		bits[j][i] = bits[i][ height-1-j];
		bits[ height-1-j][i] = buf;
	  }
	}
  }

}
