/*------------------------------------------------------
	Library for Document Analysis System
	(Text-line object class)
		Written  by H.Goto, May  1996
		Modified by H.Goto, Feb. 1997
		E-Mail: hgot@ecip.tohoku.ac.jp
------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1996-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#include	<stdio.h>
#include	<string.h>

#include	<siplib.h>
#include	"TLobj.h"


#define		RECT_ALLOC_STEP	64
#define		TL_ALLOC_STEP	16




/*---------------------------------------
	Text-line object
---------------------------------------*/

TLOBJ :: TLOBJ(){
	stat = TLSTAT_VOID;
	maxrects = rects = 0;
	rectlist = 0;
	id = attr = 0;
}


TLOBJ :: ~TLOBJ(){
	if ( rectlist != 0 )  delete []rectlist;
}


int TLOBJ :: realloc(int rects){
	TLrect	*recttmp;
	int	size;
	size = ((rects + RECT_ALLOC_STEP -1) / RECT_ALLOC_STEP) * RECT_ALLOC_STEP;
	if ( maxrects >= size )  return(0);
	if ( 0 == (recttmp = new TLrect[size]) )  return(-1);
	maxrects = size;
	memcpy((void *)recttmp,(void *)rectlist,sizeof(TLrect) * TLOBJ::rects);
	delete []rectlist;
	rectlist = recttmp;
	return(0);
}


int TLOBJ :: addrects(int rects,TLrect *rectlist){
	int	total;
	total = TLOBJ::rects + rects;
	if ( maxrects < total ){
		if ( realloc(total) )  return(-1);
	}
	memcpy((void *)(TLOBJ::rectlist + TLOBJ::rects), \
		(void *)rectlist,sizeof(TLrect) * rects);
	TLOBJ::rects += rects;
	stat |= TLSTAT_MODIFIED;
	return( TLOBJ::rects );
}


int TLOBJ :: deleteallrects(){
	if ( rectlist != 0 )  delete []rectlist;
	rectlist = 0;
	stat = TLSTAT_VOID;
	maxrects = rects = 0;
	return(0);
}


int TLOBJ :: sortrect(int align){
	int	i,j;
	TLrect	tmprect;
	if ( rectlist == 0 )  return(0);
				/* Not so good algorithm ... */
	for ( i=0 ; i < rects-1 ; i++ ){
		switch (align){
		    case 0:	/* Left end */
			for ( j=i+1 ; j<rects ; j++ ){
				if ( rectlist[i].x > rectlist[j].x ){
					tmprect = rectlist[i];
					rectlist[i] = rectlist[j];
					rectlist[j] = tmprect;
				}
			}
			break;
		    case 2:	/* Right end */
			for ( j=i+1 ; j<rects ; j++ ){
				if ( rectlist[i].x + rectlist[i].w \
				   > rectlist[j].x + rectlist[j].w ){
					tmprect = rectlist[i];
					rectlist[i] = rectlist[j];
					rectlist[j] = tmprect;
				}
			}
			break;
		    default:	/* Center */
			for ( j=i+1 ; j<rects ; j++ ){
				if ( rectlist[i].x + (rectlist[i].w /2) \
				   > rectlist[j].x + (rectlist[j].w /2) ){
					tmprect = rectlist[i];
					rectlist[i] = rectlist[j];
					rectlist[j] = tmprect;
				}
			}
			break;
		}
	}
	return(0);
}




/*---------------------------------------
	Text-line group
---------------------------------------*/

TLGroup :: TLGroup(){
	stat = TLGSTAT_VOID;
	maxlines = lines = 0;
	tllist = 0;
	id = attr = 0;
}


TLGroup :: ~TLGroup(){
	deletealllines();
}


int TLGroup :: realloc(int lines){
	TLOBJ	**tltmp;
	int	size;
	size = ((lines + TL_ALLOC_STEP -1) / TL_ALLOC_STEP) * TL_ALLOC_STEP;
	if ( maxlines >= size )  return(0);
	if ( 0 == (tltmp = new TLOBJ * [size]) )  return(-1);
	maxlines = size;
	memcpy((void *)tltmp,(void *)tllist,sizeof(TLrect *) * TLGroup::lines);
	delete []tllist;
	tllist = tltmp;
	return(0);
}


int TLGroup :: addline(TLOBJ *tlobj){
	if ( maxlines < lines + 1 ){
		if ( realloc(lines + 1) )  return(-1);
	}
	tllist[lines++] = tlobj;
	stat |= TLGSTAT_MODIFIED;
	return(lines);
}


int TLGroup :: removeline(int line){
	if ( line >= lines )  return(-1);
	if ( tllist[line] == 0 )  return(0);
	tllist[line] = 0;		/* Hole isn't packed. */
	stat |= TLGSTAT_MODIFIED;
	return(lines);
}


int TLGroup :: deletealllines(){
	int	i;
	if ( tllist != 0 ){
		for ( i=0 ; i<lines ; i++ ){
			if ( tllist[i] != 0 )  delete tllist[i];
		}
	}
	maxlines = lines = 0;
	delete []tllist;
	tllist = 0;
	stat |= TLGSTAT_MODIFIED;
	return(0);
}


int TLGroup :: removealllines(){
	int	i;
	for ( i=0 ; i<lines ; i++ )  tllist[i] = 0;
	maxlines = lines = 0;
	delete []tllist;
	tllist = 0;
	stat |= TLGSTAT_MODIFIED;
	return(0);
}


int TLGroup :: numbering(int start){
	int	i;
	TLOBJ	*tlobj;
	for ( i=0 ; i<lines ; i++ ){
		tlobj = tllist[i];
		if ( tlobj == 0 )  continue;
		tlobj->id = start++;
	}
	return(0);
}




