/*--------------------------------------------------------------
	Text line image filter            tlfilter     v1.21
	  (superficial module --- i.e. not examined carefully)
		v1.0 (Rev.961204)  by H.Goto
		v1.1 (Rev.961205)  by H.Goto
		v1.2 (Rev.970814)  by H.Goto
		v1.21(Rev.030728)  by H.Goto
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1996-2003  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of modification is specified in cases
  where modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	<siplib.h>

extern "C" int	tlimage_filter(SIPImage *tlimage);


static double	*den;
static long	*cden,*rden;
static int	cols,rows;



/*------------------------------------------------------
	Get real length of image
------------------------------------------------------*/

static int reallength(long *den,int len){
	int	s,e,i;
	for ( i=0 ; i<len ; i++ )  if ( den[i] != 0 )  break;
	s = i;
	for ( i=(len -1) ; i>=0 ; i-- )  if ( den[i] != 0 )  break;
	e = i;
	if ( 0 > (e = e - s +1) )  e = 0;
	return(e);
}




/*------------------------------------------------------
	Filter out broad black lines
------------------------------------------------------*/

#define	VRTRUN_RATIO		0.9
#define	VRTRUN_COUNTRATIO	0.15
#define	VRTRUN_DIFFRATIO	2.0

static int filter_broadline(uchar **image){
	int	i,j,u,l,c;
	int	retcode;

	for ( i=0 ; i<cols ; i++ ){
		for ( j=0 ; j<rows ; j++ )  if ( image[j][i] == 0 )  break;
		u = j;
		for ( j = (rows-1) ; j>=0 ; j-- )  if ( image[j][i] == 0 )  break;
		l = j;
		if ( 0 < (l = l - u + 1) )  den[i] = (double)cden[i] / (double)l; 
		else  den[i] = 1.0;
	}

	for ( i=0 ; i < (cols -1) ; i++ ){
		if ( cden[i] == 0 || cden[i+1] == 0 )  continue;
		if ( (double)cden[i] * VRTRUN_DIFFRATIO < (double)cden[i+1] )  den[i] = 0;
		if ( (double)cden[i+1] * VRTRUN_DIFFRATIO < (double)cden[i] )  den[i] = 0;
	}

	for ( c=i=0 ; i < (cols -1) ; i++ ){
		if ( den[i] < VRTRUN_RATIO && den[i+1] < VRTRUN_RATIO )  c++;
	}

	if ( c >= (int)(VRTRUN_COUNTRATIO * (double)rows) )
		retcode = 1;		/* Text Line */
	else    retcode = 0;		/* Not Text Line */

	return(retcode);
}




/*------------------------------------------------------
	Filter out dithered image
------------------------------------------------------*/

#define	CROSSING_LOCALLIMIT	8
#define	CROSSING_GLOBALLIMIT	4

static int filter_dither(uchar **image){
	int	i,j,c,a,s,ew;
	int	*edges;
	int	retcode = 1;
	if ( rows == 0 )  return(0);
	if ( 0 == (edges = new int[cols]) )  return(-2);
	for ( a=c=i=0 ; i < (cols -1) ; i++ )
		if ( cden[i] == 0 && cden[i+1] != 0 ){
			edges[i] = 1;  ++a;
		}
		else	edges[i] = 0;
	edges[cols -1] = 0;
	s = reallength(rden,rows);	/* Assume character width */
	if ( s == 0 )  return(0);
	if ( a > CROSSING_GLOBALLIMIT * s ){
		retcode = 0;
	}
	else{	for ( i=0 ; i<cols ; i++ ){
			if ( cols < (ew = i + s) )  ew = cols;
			for ( c=0, j=i ; j<ew ; j++ )  c += edges[j];
			if ( c > CROSSING_LOCALLIMIT ){ retcode = 0; break; }
		}
	}
	delete []edges;
	return(retcode);
}




/*------------------------------------------------------
	Filter out short regions
------------------------------------------------------*/

#define	ASPECT_LIMIT	1.2

static int filter_short(uchar **image){
	int	l,h;
	l = reallength(cden,cols);
	h = reallength(rden,rows);
	if ( (double)l < (double)h * ASPECT_LIMIT )  return(0);
	return(1);
}




/*------------------------------------------------------
	Check if the image is a true text line
------------------------------------------------------*/

int tlimage_filter(SIPImage *tlimage){
	int	i,j;
	uchar	**image;
	int	retcode = 0;

	cols = tlimage->width;
	rows = tlimage->height;

	if ( 0 == (image = new uchar* [rows]) )  return(-2);
	if ( 0 == (den = new double[cols]) ){
		delete []image;
		return(-2);
	}
	if ( 0 == (cden = new long[cols + rows]) ){
		delete []image;
		delete []den;
		return(-2);
	}
	rden = cden + cols;

	for ( j=0 ; j<rows ; j++ )  image[j] = (uchar *)sip_getimgptr(tlimage,j);
		/* Assume  image[j][i] = 0: black pixel, 255: white pixel */ 

	sip_projprofile(tlimage,rden,cden);
	for ( i=0 ; i<cols ; i++ )  cden[i] = (long)rows - (cden[i]/255);
	for ( j=0 ; j<rows ; j++ )  rden[j] = (long)cols - (rden[j]/255);

	if ( 1 == (retcode = filter_broadline(image)) )
	if ( 1 == (retcode = filter_dither(image)) )
	if ( 1 == (retcode = filter_short(image)) )		;


	delete []cden;
	delete []image;
	delete []den;

	return(retcode);
}


